<?php

/**
 * Tinkoff Bank Payment package for Concrete CMS Community Store
 *
 * The MIT License (MIT)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:

 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.

 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 * 
 * @author Alex Borisov <linuxoidoz@gmail.com>
 * @copyright 2021-2022, Alex Borisov
 * @package Concrete\Package\community_store_tinkoff
 */

namespace Concrete\Package\CommunityStoreTinkoff\Src\CommunityStore\Payment\Methods\CommunityStoreTinkoff;

use Concrete\Core\Support\Facade\Application;
use Concrete\Core\Support\Facade\Config;
use Concrete\Core\Support\Facade\Session;
use Concrete\Core\Support\Facade\Log;
use Concrete\Core\Support\Facade\Url;
use Concrete\Core\Support\Facade\DatabaseORM as ORM;
use Concrete\Core\Routing\Redirect;
use Concrete\Core\Http\Request;
use Symfony\Component\HttpFoundation\IpUtils;
use GuzzleHttp\Client as GuzzleHttpClient;
use GuzzleHttp\RequestOptions as GuzzleHttpClientRequestOptions;

use Concrete\Package\CommunityStore\Src\CommunityStore\Payment\Method as StorePaymentMethod;
use Concrete\Package\CommunityStore\Src\CommunityStore\Order\Order as StoreOrder;
use Concrete\Package\CommunityStore\Src\CommunityStore\Order\OrderStatus\OrderStatus as StoreOrderStatus;

class CommunityStoreTinkoffPaymentMethod extends StorePaymentMethod
{
    public function dashboardForm()
    {
        $this->set('tinkoffTerminalKey', Config::get('community_store_tinkoff.terminalKey'));
        $this->set('tinkoffTerminalPassword', Config::get('community_store_tinkoff.terminalPassword'));
        
        $this->set('testTerminalKey', Config::get('community_store_tinkoff.testTerminalKey'));
        $this->set('testTerminalPassword', Config::get('community_store_tinkoff.testTerminalPassword'));
        
        $this->set('tinkoffTax', Config::get('community_store_tinkoff.tax'));
        $this->set('tinkoffTaxRate', Config::get('community_store_tinkoff.rate'));
        $this->set('tinkoffPaymentObject', Config::get('community_store_tinkoff.object'));
        $this->set('tinkoffLanguage', Config::get('community_store_tinkoff.language'));
        $this->set('tinkoffTestMode', Config::get('community_store_tinkoff.testMode'));
        
        $languages = [
            'en' => t('English'), 
            'ru' => t('Russian')
        ];
        
        $this->set('languages', $languages);
        
        $taxes = [
            'osn' => t('General'),
            'usn_income' => t('Simplified (income)'),
            'usn_income_outcome' => t('Simplified (income minus expenses)'),
            'envd' => t('Single tax on imputed income'),
            'esn' => t('Uniform agricultural'),
            'patent' => t('Patent'),
        ];
        
        $this->set('taxes', $taxes);
        
        $tax_rates = [
            'none' => t('None'),
            'vat0' => '0 %',
            'vat10' => '10 %',
            'vat18' => '18 %',
            'vat20' => '20 %',
        ];
        
        $this->set('tax_rates', $tax_rates);
        
        $payment_objects = [
            'commodity' => t('Commodity'),
            'excise' => t('Excise'),
            'job' => t('Job'),
            'service' => t('Service'),
        ];
        
        $this->set('payment_objects', $payment_objects);
        
        $app = Application::getFacadeApplication();
        $this->set('form', $app->make('helper/form'));
        
        $db = $app->make('database')->connection();
        $q = 'SELECT COUNT(*) AS num FROM CommunityStorePaymentTinkoff';
        $log_count = $db->fetchAssoc($q);
        $this->set('log_count', $log_count['num'] ?: 0);
    }

    public function save($data = array())
    {
        Config::save('community_store_tinkoff.terminalKey', $data['tinkoffTerminalKey'] ? trim($data['tinkoffTerminalKey']) : '');
        Config::save('community_store_tinkoff.terminalPassword', $data['tinkoffTerminalPassword'] ? trim($data['tinkoffTerminalPassword']) : '');
        
        Config::save('community_store_tinkoff.testTerminalKey', $data['testTerminalKey'] ? trim($data['testTerminalKey']) : '');
        Config::save('community_store_tinkoff.testTerminalPassword', $data['testTerminalPassword'] ? trim($data['testTerminalPassword']) : '');
        
        Config::save('community_store_tinkoff.tax', $data['tinkoffTax'] ?: '');
        Config::save('community_store_tinkoff.rate', $data['tinkoffTaxRate'] ?: '');
        Config::save('community_store_tinkoff.object', $data['tinkoffPaymentObject'] ?: '');
        Config::save('community_store_tinkoff.language', $data['tinkoffLanguage'] ?: '');
        Config::save('community_store_tinkoff.testMode', $data['tinkoffTestMode'] ? 1 : 0);
    }

    public function validate($data, $e)
    {
        $pm = StorePaymentMethod::getByHandle('community_store_tinkoff');
        
        if ($data['paymentMethodEnabled'][$pm->getID()] == 1) {
            if ($data['tinkoffTestMode']) {
                if (empty($data['testTerminalKey'])) {
                    $e->add(t('Test Terminal Key must be set'));
                }

                if (empty($data['testTerminalPassword'])) {
                    $e->add(t('Test Terminal Password must be set'));
                }
            }
            else {
                if (empty($data['tinkoffTerminalKey'])) {
                    $e->add(t('Tinkoff Terminal Key must be set'));
                }

                if (empty($data['tinkoffTerminalPassword'])) {
                    $e->add(t('Tinkoff Terminal Password must be set'));
                }
            }
            
            if ($data['currency'] !== 'RUB') {
                $e->add(t('Tinkoff payment can only be made in Rubles. Please set Russian Ruble as store currency in Store Settings.'));
            }
        }
        return $e;
    }

    public function redirectForm()
    {
        $order = StoreOrder::getByID(Session::get('orderID'));
        $gateway = $this->getGatewayURL($order);

        if ($order && !empty($gateway['payment_id'])) {
            $order->setTransactionReference($gateway['payment_id']);
            $order->save();
        }
        
        $r = Redirect::to($gateway['link']);
        $r->send();
        exit;
    }

    public static function validateCompletion()
    {
        // if we get to this function, we returned from tinkoff after a payment
        $r = Redirect::to('/checkout/complete');
        $r->send();
        exit;

    }

    public function checkoutForm()
    {
        $pmID = StorePaymentMethod::getByHandle('community_store_tinkoff')->getID();
        $this->set('pmID', $pmID);
    }

    private static function getGatewayURL($order)
    {
        $success = (string)Url::to('/checkout/tinkoffresponse');
        $fail = (string)Url::to('/checkout/failed#payment');
        $notify = (string)Url::to('/checkout/tinkoffcallback');
        
        if (Config::get('community_store.currency') !== 'RUB') {
            Session::set('paymentErrors', t('Incorrect currency, please contact the site administrator.'));
            return ['payment_id' => '', 'link' => $fail];
        }
        
        $terminalKey = Config::get('community_store_tinkoff.terminalKey');
        $language = Config::get('community_store_tinkoff.language');
        $tax = Config::get('community_store_tinkoff.tax');
        $tax_rate = Config::get('community_store_tinkoff.rate');
        $payment_object = Config::get('community_store_tinkoff.object');
        $testMode = Config::get('community_store_tinkoff.testMode');
        
        $order_id = $order->getOrderID();
        
        if ($testMode) {
            $terminalKey = Config::get('community_store_tinkoff.testTerminalKey');
            
            $order_id = 'CSTEST' . $order_id;
            $url = 'https://rest-api-test.tinkoff.ru/v2/Init/';
        }
        else {
            $url = 'https://securepay.tinkoff.ru/v2/Init/';
        }
        
        if ($order_id) {
            $description = t('Order #%s', $order_id);
            
            $value = number_format($order->getTotal(), 2, '.', '') * 100;
            
            try {
                $orderItems = $order->getOrderItems();
                $items = array();
                $total = 0;
                
                foreach ($orderItems as $item) {
                    $price = round($item->getPricePaid() * 100);
                    $quantity = $item->getQuantity();
                    $amount = round($price * $quantity);
                    
                    $item = [
                        'Name' => $item->getProductName(),
                        'Price' => $price,
                        'Quantity' => $quantity,
                        'Amount' => $amount,
                        'PaymentMethod' => 'full_prepayment',
                        'PaymentObject' => $payment_object,
                        'Tax' => $tax_rate,
                    ];
                    
                    array_push($items, $item);
                    $total += $amount;
                }
                
                $receipt = [
                    'Email' => $order->getAttribute('email'),
                    'Phone' => $order->getAttribute('billing_phone'),
                    'Taxation' => $tax,
                    'Items' => $items,
                ];
                
                $token_data = [
                    'TerminalKey' => $terminalKey,
                    'Amount' => $value,
                    'OrderId' => $order_id,
                    'Description' => $description,
                ];
                
                $token = self::getToken($token_data);
                
                $data = [
                    'TerminalKey' => $terminalKey,
                    'Amount' => $value,
                    'OrderId' => $order_id,
                    'Description' => $description,
                    //'Token' => $token,
                    'Language' => $language,
                    'PayType' => 'O',
                    'SuccessURL' => $success,
                    'FailURL' => $fail,
                    'NotificationURL' => $notify,
                    'Receipt' => $receipt,
                    'DATA' => [
                        'Email' => $order->getAttribute('email'),
                        'Phone' => $order->getAttribute('billing_phone'),
                    ],
                ];
                
                $options = array(
                    'http' => array(
                        'method' => 'POST',
                        'content' => json_encode($data),
                        'header' =>  "Content-Type: application/json\r\n" . "Accept: application/json\r\n",
                    )
                );
                
                $app = Application::getFacadeApplication();
                $db = $app->make('database')->connection();
                $dbTable = 'CommunityStorePaymentTinkoff';
                
                if ($db->tableExists($dbTable)) {
                    $ptOrder = $testMode ? str_replace('CSTEST', '', $order_id) : $order_id;
                    
                    $q = 'INSERT INTO ' . $dbTable . ' (ptOrder, ptTotal, ptToken, ptTerminal) VALUES(?,?,?,?) ON DUPLICATE KEY UPDATE ptOrder = ?';
                    $v = [(int)$ptOrder, (int)$total, $token, $terminalKey, (int)$ptOrder];
                    $db->executeQuery($q, $v);
                }
                
                if (version_compare(Config::get('concrete.version'), '9.0', '<')) {
                    $context  = stream_context_create($options);
                    $result = file_get_contents($url, false, $context);
                    $response = json_decode($result);
                }
                else {
                    $client = $app->make(GuzzleHttpClient::class);
                    $response = $client->post($url, [
                        GuzzleHttpClientRequestOptions::JSON => $data
                    ]);
                    $response = json_decode($response->getBody()->getContents());
                }
                
                if (!empty($response) && $response->TerminalKey === $terminalKey && $response->Success === true && $response->Status === 'NEW') {
                    $payment_id = $response->PaymentId ?: 0;
                    $link = $response->PaymentURL ?: $fail;
                    return ['payment_id' => $payment_id, 'link' => $link];
                }
                else {
                    Session::set('paymentErrors', !empty($response->Message) ? $response->Message . '. ' . $response->Details : t('Payment error. Please try again'));
                    return ['payment_id' => '', 'link' => $fail];
                }
            }
            catch (\Exception $e) {
                Session::set('paymentErrors', $e->getMessage());
                return ['payment_id' => '', 'link' => $fail];
            }
        }
        else {
            return ['payment_id' => '', 'link' => $fail];
        }
    }

    public static function callback()
    {        
        $success = false;
        $request = json_decode(file_get_contents('php://input'));
        $request->Success = $request->Success ? 'true' : 'false';
        $request = (array)$request;
        Log::Info(var_dump_safe($request));
        
        if (empty($request)) {
            exit;
        }
        Log::Info('request ok'); //******************************************************* DELETE **************************************************************
        
        $tinkoff_token = $request['Token'];
        
        $token_data = [
            'TerminalKey' => $request['TerminalKey'],
            'OrderId' => $request['OrderId'],
            'Success' => $request['Success'],
            'Status' => $request['Status'],
            'PaymentId' => $request['PaymentId'],
            'ErrorCode' => $request['ErrorCode'],
            'Amount' => $request['Amount'],
            'CardId' => $request['CardId'],
            'Pan' => $request['Pan'],
            'ExpDate' => $request['ExpDate'],
        ];
        
        $token = self::getToken($token_data);
        
        if ($token !== $tinkoff_token) {
            exit;
        }
        Log::Info('token ok'); //******************************************************* DELETE **************************************************************
        
        $app = Application::getFacadeApplication();
        $r = $app->make(Request::class);
        $ip = $r->getClientIp();
        
        if (!IpUtils::checkIp($ip, '91.194.226.0/23')) {
            exit;
        }
        Log::Info('ip ok'); //******************************************************* DELETE **************************************************************
        
        $order_id = $request['OrderId'];
        $amount = $request['Amount'];
        $terminalKey = $request['TerminalKey'];
        $testMode = Config::get('community_store_tinkoff.testMode');
        
        if ($testMode) {
            $order_id = str_replace('CSTEST', '', $order_id);
        }
            
        $dbTable = 'CommunityStorePaymentTinkoff';
        $db = $app->make('database')->connection();
        
        if ($db->tableExists($dbTable)) {
            $q = 'SELECT * FROM ' . $dbTable . ' WHERE ptOrder = ?';
            $v = [(int)$order_id];
            $row = $db->fetchAssoc($q, $v);
            
            if (empty($row)) {
                exit;
            }
            Log::Info('order ok'); //******************************************************* DELETE **************************************************************
            
            if ((int)$row['ptTotal'] !== (int)$amount) {
                exit;
            }
            Log::Info('total ok'); //******************************************************* DELETE **************************************************************
            
            if ($row['ptTerminal'] !== $terminalKey) {
                exit;
            }
            Log::Info('terminal ok'); //******************************************************* DELETE **************************************************************
        }
        
        $status = strtoupper($request['Status']);
        $payment_id = $request['PaymentId'];

        if ($status === 'CONFIRMED' ) {
            $order = StoreOrder::getByID($order_id);
            
            if ($order) {
                if ($order->getExternalPaymentRequested()) {
                    $order->completeOrder($payment_id);
                    $order->updateStatus(StoreOrderStatus::getStartingStatus()->getHandle());
                    $success = true;
                }
            }
        }
        elseif ($status === 'CANCELED' || $status === 'REVERSED' || $status === 'REJECTED') {
            $order->setPaid(null);
            $order->save();
        }
        elseif ($status === 'REFUNDED') {
            $em = ORM::entityManager();
            $order = $em->getRepository('Concrete\Package\CommunityStore\Src\CommunityStore\Order\Order')->findOneBy(array('transactionReference' => $payment_id));
            
            if ($order) {
                $order->setRefunded(new \DateTime());
                //$order->setRefundReason(t('REFUNDED'));
                $order->save();
                $success = true;
            }                
        }
        
        if ($success) {
            http_response_code(200);
            exit('OK');
        }
        else {
            http_response_code(400);
        }
        
        exit;
    }

    public static function getToken($data)
    {
        $data['Password'] = Config::get('community_store_tinkoff.terminalPassword');
        $testMode = Config::get('community_store_tinkoff.testMode');
        
        if ($testMode) {
            $data['Password'] = Config::get('community_store_tinkoff.testTerminalPassword');
        }
        
        unset($data['Token']);
        ksort($data);
        $values = implode('', array_values($data));
        return hash('sha256', $values);
    }
    
    public static function clearlog()
    {
        $app = Application::getFacadeApplication();
        $db = $app->make('database')->connection();
        $table = 'CommunityStorePaymentTinkoff';
        
        if ($db->tableExists($table)) {
            $db->executeQuery('DELETE FROM ' . $table);
        }
        
        $r = Redirect::to('dashboard/store/settings#settings-payments');
        $r->send();
        exit;
    }

    public function getPaymentMinimum()
    {
        return 1.0;
    }

    public function getName()
    {
        return t('Tinkoff');
    }

    public function isExternal()
    {
        return true;
    }




}
